package com.appiancorp.core.expr;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import javax.xml.namespace.QName;

import com.google.common.annotations.GwtCompatible;

/**
 *
 * THIS IS THE GWT EMULATED COPY OF THIS CLASS.
 *
 * Although it is in sync with the ae copy, you should not be using this code.
 *
 * Domain is the enumeration of valid Id Domains.
 *
 * They are all proper folded form already, so they need not be refolded.
 *
 * See backend EVD_name from pcc.k.
 *
 * @author brian.sullivan
 */
@GwtCompatible
public class Domain implements Comparable {

  // MAINT: Map MUST be defined before the static constants, as they're also defined in the static initializer context
  private static final ConcurrentHashMap<String, Domain> map = new ConcurrentHashMap<String, Domain>();

  // Domain Constants; these may be treated as enum constants, used directly (as they were originally enum constants before custom domains)
  public static final Domain DEFAULT = new Domain(null, null, 0, true);
  public static final Domain CT = new Domain("*", null, 1);
  public static final Domain PM = new Domain("pm", "Process Model Properties", 2);
  public static final Domain PV = new Domain("pv", "Process Variables", 3);
  public static final Domain PP = new Domain("pp", "Process Properties", 4);
  public static final Domain TP = new Domain("tp", "Task Properties", 5);
  public static final Domain AC = new Domain("ac", "Node Inputs or Outputs", 6);
  public static final Domain FE = new Domain("fe", null, 7);
  public static final Domain MSG = new Domain("msg", "Message Properties", 8);
  public static final Domain PVP = new Domain("pvp", null, 9);
  public static final Domain RULE = new Domain("rule", "Rule(s)", 10, false, false, true, false);  // rule
  public static final Domain CONS = new Domain("cons", "Constant(s)", 11, true);     // constant
  public static final Domain RI = new Domain("ri", "Rule Input", 12);
  public static final Domain BIND = new Domain("bind", null, 14);
  public static final Domain ENTRY = new Domain("entry", null, 15);

  public static final Domain FN = new Domain("fn", null, 17, false, true, true, true);            // function
  public static final Domain EXT = new Domain("ext", null, 18, false, true, false, true);         // external
  public static final Domain INT = new Domain("int", null, 19, false, true, false, true);         // internal
  public static final Domain TYPE = new Domain("type", null, 20, false, false, true, true);       // type constructors
  public static final Domain CUSTOM = new Domain("custom", null, 21, false, false, false, true); // custom function
  public static final Domain DATA = new Domain("data", null, 22);     // data context

  public static final Domain PARSE = new Domain("parse", null, 23); // parse context, used for reflective information about the parsing
  public static final Domain LOCAL = new Domain("local", "Local Variables", 24);

  public static final Domain RECORD_FIELD = new Domain("rf", "Record Fields", 25);
  public static final Domain RECORD_PROPERTY = new Domain("rp", "Record Properties", 26);
  public static final Domain RECORD_TYPE_PROPERTY = new Domain("rt", "Record Type Properties", 27);
  public static final Domain RECORD = new Domain("record", "Record", 28);
  public static final Domain SYS = new Domain("a", "System Rule(s)", 29, false, false, true, false);  // rule
  public static final Domain RECORD_SOURCE_PARAMETER = new Domain("rsp", "Record Source Parameters", 32);

  public static final Domain FLOW = new Domain("flow", "SAIL Execution Flow", 30);
  public static final Domain LAMBDA = new Domain("lambda", "Lambdas", 31, false, true, false, false);  // id reference for dynamically generated lambdas
  public static final Domain SASA = new Domain("sasa", "SASA Variables", 33);
  public static final Domain SAVE = new Domain("save", "Saves Implicit Binding(s)", 34);

  private String domainName;
  private boolean constant;
  private boolean eval;
  private String longDomainName;
  private boolean reference;  // indicates that this variable need not be found initially as can represent a reference type
  private boolean dotRetained;
  private boolean custom;
  private boolean standard;

  private Domain(String domainName, String longDomainName, int domainId) {
    this(domainName, longDomainName, domainId, false);
  }

  private Domain(String domainName, String longDomainName, int domainId, boolean constant) {
    this(domainName, longDomainName, domainId, constant, false, false, false);
  }

  private Domain(String domainName, String longDomainName, int domainId, boolean constant, boolean eval, boolean reference, boolean dotRetained) {
    this(domainName, longDomainName, domainId, constant, eval, reference, dotRetained, false, true);
  }

  private Domain(String domainName, String longDomainName, int domainId, boolean constant, boolean eval, boolean reference, boolean dotRetained, boolean custom, boolean standard) {
    this.domainName = domainName != null ? domainName.toLowerCase() : null;
    this.longDomainName = longDomainName;
    this.constant = constant;
    this.eval = eval;
    this.reference = reference;
    this.dotRetained = dotRetained;
    this.custom = custom;
    this.standard = standard;
    if(domainName != null) {
      map.putIfAbsent(domainName, this);
    }
  }

  /*
   * Custom Domain
   */
  private Domain(String domainName) {
    this(domainName, null, -1, false, false, false, false, true, false);
  }

  public String getDomainName() {
    return domainName;
  }

  public boolean isCustom() {
    return custom;
  }

  public static Domain getDomain(String domainName) {
    if(domainName == null) {
      return DEFAULT;
    }
    final Domain domain = domainName.length()>0 ? map.get(domainName.toLowerCase()) : null;
    if(domain == null) {
      new Domain(domainName);  // construct always, but get from map again as it will only be inserted into map if not already present; any garbage Domain from race condition will be automatically collected.
      return map.get(domainName);
    }
    return domain;
  }

  public String getExpression(String name){
    return this.domainName + Id.SEPARATOR + name.toLowerCase();
  }

  public String getCasedExpression(String name){
    return this.domainName + Id.SEPARATOR + name;
  }

  public boolean isDomain(Domain d) {
    return (d == null) ? false : d == this;
  }

  /**
   * Filter to only those ids in this Domain.
   *
   * @param ids
   * @return
   */
  public List<Id> filter(Collection<Id> ids) {
    final List<Id> results = new ArrayList<Id>();
    final Iterator<Id> iter = ids.iterator();
    while(iter.hasNext()) {
      Id id = iter.next();
      if(isDomain(id.getDomain())) {
        results.add(id);
      }
    }
    return results;
  }

  /**
   * Filter to only those ids in this Domain, using those ids' original keys as the source of QName.valueOf.
   *
   * Particularly useful for domain literals, e.g., type!{ns}name.
   *
   * @param ids
   * @return
   */
  public List<QName> filterOriginalKeyToQName(Collection<Id> ids) {
    final List<Id> filtered = filter(ids);
    final List<QName> results = new ArrayList<QName>();
    for(Id id:filtered) {
      results.add(QName.valueOf(id.getOriginalKey()));
    }
    return results;
  }

  /**
   * Can this domain possibly refer to a constant?
   *
   * @return
   */
  public boolean isConstant() {
    return constant;
  }

  /**
   * Other than as a rule, can this domain be evaluated?
   *
   * @return
   */
  public boolean isEval() {
    return eval;
  }

  /**
   * Suitable for error messages.
   *
   * @return
   */
  public String getLongDomainName() {
    return longDomainName;
  }

  /**
   * Can this domain represent a reference?
   *
   * @return
   */
  public boolean isReference() {
    return reference;
  }

  /**
   * Is dot retained even without single quotes?  Generally true for functions, false for domains of data that can be drilled.
   *
   * @return true if retained
   */
  public boolean isDotRetained() {
    return dotRetained;
  }

  /**
   * @return domain name
   */
  @Override
  public String toString() {
    return domainName;
  }

  /**
   * Standard domains are those built-in, not arbitrarily defined.
   *
   * @return
   */
  public boolean isStandard() {
    return standard;
  }

  /**
   * Compare based on Domain name.
   */
  @Override
  public int compareTo(Object o) {
    if(o == null || !(o instanceof Domain)) {
      return -1;
    }
    final String otherDomainName = ((Domain)o).domainName;
    return domainName == null ? (otherDomainName == null ? 0 : -1) : otherDomainName != null ? domainName.compareTo(otherDomainName) : -1;
  }
}