package com.appiancorp.core.expr;

import com.appiancorp.suiteapi.common.DeepCloneable;
import com.google.common.annotations.GwtCompatible;

/**
 * THIS IS THE GWT EMULATED COPY OF THIS CLASS.
 *
 * Although it is in sync with the ae copy, you should not be using this code.
 *
 * An {@link Id} is an immutable, case-insensitive, qualified (i.e. domain!key) identifier. When the
 * domain is elided, it is in the default domain, which can be bound to different domains at different
 * times, such as {@link Domain#PV} for a process evaluation context, {@link Domain#RI} for rule
 * evaluation context, etc.
 *
 * @author brian.sullivan
 */
@GwtCompatible
public class Id implements DeepCloneable, Comparable {

  /**
   * The SEPARATOR separates the domain from the key in the name.
   *
   * (This corresponds to sheet!cell in a spreadsheet.)
   */
  public static final char SEPARATOR = '!';

  private static final char DOT = '.';

  private final Domain domain;    // domain will be null or lowercase domain, never empty

  private final String key;       // key will never be null; it will be lowercase
  private final String okey;      // original key, will never be null; it will not be lowercased

  private String name;      // name is lazily set, not generated until required; it will always be lowercase
  private String oName;     // original name, not lowercased

  private boolean nameSet;  // nameSet is true after name has been lazily set

  private final boolean dot;      // name begins with a dot (.), so it's actually a property
  private final boolean simple;   // if name can be output as expression without single quotes

  /**
   * Construct an identifier from a full name, e.g., ("pv!myvariable")
   *
   * @param name
   */
  public Id(String name) {
    this(name, false);
  }

  /**
   * Construct an identifier from an identifier in another domain.
   *
   * Typically used to reset the default domain to known domain.
   *
   * @param base
   * @param alterDomain
   */
  public Id(Id base, Domain alterDomain) {
    this.domain = alterDomain;

    this.key = base.key;
    this.okey = base.okey;

    this.dot = base.dot;

    resetName();  // force re-evaluation of the name
    String plainDomainString = domain == null || domain == Domain.DEFAULT ? "" : domain.getDomainName();
    simple = isSimple(plainDomainString, false) && isSimple(okey, true);
  }

  public Id(String name, boolean dot) {
    if(name == null) {
      throw new NullPointerException("Cannot create an identifier with null name");
    }
    if(name.length() == 0) {
      throw new IllegalArgumentException("Cannot create an identifier with zero length key");
    }
    this.dot = dot;
    final int index = name.indexOf('!');
    if(index>=0) {
      if(index == (name.length()-1)) {
        throw new IllegalArgumentException("Cannot create an identifier with zero length key");
      }
      // domain and key
      final String domainName = name.substring(0, index);
      okey = name.substring(index+1);
      domain = domainName.length() == 0 ? Domain.DEFAULT : Domain.getDomain(domainName);
    } else {
      okey = name;
        domain = Domain.DEFAULT;
      }
    if(okey.length() == 0) {
        throw new IllegalArgumentException("Cannot create an identifier with zero length key");
      }
    key = okey.toLowerCase();
    this.oName = name;
    this.name = name.toLowerCase();
    nameSet = true;

    String plainDomainString = domain == null || domain == Domain.DEFAULT ? "" : domain.getDomainName();
    simple = isSimple(plainDomainString, false) && isSimple(okey, true);
  }

  /**
   * Construct an identifier from separate domain and key, e.g., ("pv","myVariable")
   * @param domain
   * @param key
   */
  public Id(String domainName, String key) {
    this(domainName, key, false);

  }
  public Id(String domainName, String key, boolean dot) {
    if(key == null) {
      throw new NullPointerException("Cannot create an identifier with null key");
    }
    if(key.length() == 0) {
      throw new IllegalArgumentException("Cannot create an identifier with zero length key");
    }
    this.dot = dot;
    domainName = domainName != null ? (domainName.length()>0 ? domainName.toLowerCase() : null) : domainName;
    domain = Domain.getDomain(domainName);
    this.okey = key;
    this.key = key.toLowerCase();

    String plainDomainString = domain == null || domain == Domain.DEFAULT ? "" : domain.getDomainName();
    simple = isSimple(plainDomainString, false) && isSimple(okey, true);
  }

  public Id(Domain domain, String key) {
    this(domain, key, false);
  }

  public Id(Domain domain, String key, boolean dot) {
    if(key == null) {
      throw new NullPointerException("Cannot create an identifier with null key");
    }
    if(key.length() == 0) {
      throw new IllegalArgumentException("Cannot create an identifier with zero length key");
    }
    this.dot = dot;
    this.domain = domain;
    this.okey = key;
    this.key = key.toLowerCase();

    String plainDomainString = domain == null || domain == Domain.DEFAULT ? "" : domain.getDomainName();
    simple = isSimple(plainDomainString, false) && isSimple(okey, true);
  }

  private void setName() {
    if(nameSet) {
      return;
    }
    resetName();
    }

  private void resetName() {
    final String domainName = domain != null ? domain.getDomainName() : null;
    if(domainName != null && domainName.length() > 0) {
      name = domainName + SEPARATOR + key;
      oName = domainName + SEPARATOR + okey;
    } else {
      name = key;
      oName = okey;
    }
    nameSet = true;
  }

  /**
   * Original name has not been forced to lowercase.  It includes the domain if provided.
   *
   * Original Name is domain!originalKey or originalKey, but never quoted.
   * Original Name does not include dot if present.
   *
   * @return
   */
  @Deprecated
  public String getOriginalExpression() {
    setName();
    return oName;
  }

  /**
   * Domain of either null or Domain.DEFAULT is considered default.
   *
   * @return
   */
  public Domain getDomain() {
    return domain;
  }

  /**
   * Key has been forced to lowercase.  It never includes the domain.
   * Key does not include dot if present.
   *
   * @return
   */
  public String getKey() {
    return key;
  }

  /**
   * Original key has not been forced to lowercase.  It never includes the domain.
   * Original key does not include dot if present.
   *
   * @return
   */
  public String getOriginalKey() {
    return okey;
  }

  /**
   * Get the original key, removing the domain if this Id matches the given defaultDomain.
   *
   * @param defaultDomain
   * @return
   */
  public String getOriginalKey(Domain defaultDomain) {
    if(defaultDomain != null && domain != null && defaultDomain.equals(domain)) {
      return getOriginalKey();  // the key does not include the domain!; original case is preserved
    } else {
      return toString(true, true);  // the id, including domain!key, but not including single quotes
    }
  }

  /**
   * Name is domain!key or key, but never quoted.
   * Name does not include dot if present.
   *
   * @return
   */
  public String getName() {
    setName();
    return name;
  }

  public boolean isDefaultDomain() {
    return domain == null || domain == Domain.DEFAULT;
  }

  /**
   * Does this Id have either the specified domain or Domain.DEFAULT?
   *
   * @param d
   *          The domain to check against. This should usually be the domain that is the
   *          default in the context in which the Id is being used.
   * @return true if the domain is null, if it's Domain.DEFAULT, or if it's equal to the
   *         domain passed in
   */
  public boolean isDomainOrDefault(Domain domain) {
    return isDefaultDomain() || this.domain == domain;
  }

  public boolean isDomain(Domain checkDomain) {
    if(domain == null) {
      return checkDomain == null;
    }
    if(checkDomain == null) {
      return false;
    }
    return domain.isDomain(checkDomain);
  }

  @Override
  public boolean equals(Object o) {
    if(!(o instanceof Id)) {
      return false;
    }
    Id other = (Id)o;
    if(!(dot == other.dot)) {
      return false;
    }
    if(!key.equals(other.key)) {
      return false;
    }
    return dot ? true : isDomain(other.domain);  // domain only matters if not dotted
  }

  @Override
  public int hashCode() {
    if(dot) {
      int hash = 3*key.hashCode();  // skip domain if dotted
      hash^=dot ? 123 : 456;
      return hash;
    } else {
      int hash = domain!=null ? domain.hashCode() : 0;  // domain only matters if not dotted
    hash^=3*key.hashCode();
    hash^=dot ? 123 : 456;
    return hash;
  }
  }

  /**
   * This returns the original name (including original case), single quoted as appropriate, and including any property dot as appropriate.
   */
  @Override
  public String toString() {
    setName();
    return toString(simple, false);
  }

  /**
   * @param skipSingleQuotes indicates that 'quotes' will not surround the result, even if the Id would require them in an expression
   */
  public String toString(boolean skipSingleQuotes) {
    return toString(skipSingleQuotes, false);
  }

  /**
   * @param skipSingleQuotes indicates that 'quotes' will not surround the result, even if the Id would require them in an expression
   * @param simplify indicates that id simplification will occur, replacing !originalName with originalName if no domain is specified
   */
  private String toString(boolean skipSingleQuotes, boolean simplify) {
    setName();
    final String name = simplify && oName.length() > 0 && oName.charAt(0) == SEPARATOR ? oName.substring(1) : oName;
    if(dot) {
      return skipSingleQuotes ? DOT + name : DOT + "'" + name +"'"; // dotted identifiers have no domain
    } else {
      return skipSingleQuotes ? name : "'"+name +"'";
    }
  }

  private static boolean EVER_ALLOW_BRACE_NAMESPACE_IN_SIMPLE = false;  // set to true when rule save interface allows this

  public static boolean isSimple(String text) {
    return isSimple(text, false);
  }

  public static boolean isSimple(String text, boolean allowBraceNamespace) {
    if(text == null) {
      return true;
    }
    final int length = text.length();
    if(length < 1) {
      return false;
    }

    char ch = text.charAt(0);
    int start = 1;
    // can begin with {namespace}, where braces act as quotes, and still be simple
    if(allowBraceNamespace && EVER_ALLOW_BRACE_NAMESPACE_IN_SIMPLE) {
      if(ch == '{') {
        for(;start<length;start++) {
          ch = text.charAt(start);
          if(ch == '}') {
            start++;
            if(start >= length) {
              return true;
            }
            ch = text.charAt(start);
            break;
          }
        }
      }
    }
    if((ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || ch == '@') {  // initial character is more restrictive
      for(int i=start; i<length; i++) {
        ch = text.charAt(i);
        if(!((ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || (ch>='0' && ch<='9') || ch=='_' || ch == '@')) {  // subsequent characters can include digits and _
          return false;
        }
      }
      return true;
      } else {
      return false;
      }
    }

  // PORT: Rename to isProperty at the end
  public boolean isDot() {
    return dot;
  }

  @Override
  @edu.umd.cs.findbugs.annotations.SuppressWarnings(value="CN_IDIOM_NO_SUPER_CALL", justification="Ids are immutable & Projection.java overrides clone()")
  public Object clone() {
    return this;
  }

  /**
   * Compare based on domain, then key.
   */
  @Override
  public int compareTo(Object o) {
    if(o == null || !(o instanceof Id)) {
      return -1;
    }
    final Id otherId = (Id)o;
    if(domain == null) {
      return otherId.domain == null ? 0 : 1;
    }
    final int compare = domain.compareTo(otherId.domain);
    if(compare != 0) {
      return compare;
    }
    return key == null ? (otherId.key == null ? 0 : 1)
         : otherId.key != null ? key.compareTo(otherId.key)
         : -1;
  }
}